unit uEmerald;

////////////////////////////////////////////////////////////////////////////////
//
// Emerald Encryption version 6.4
//
// copyright (c) Steve Corbett 1994-1998
//
// This component can be used to encrypt/decrypt any kind of data file
// using proprietary Emerald (c) encryption.
//
// USAGE:
//   1) Supply a password.  You can supply more than one password by separating
//      them with spaces.  The more passwords you supply, the more secure the
//      encryption process (using words of differing lengths will further
//      enhance the security).
//   2) Supply a source file name.  This can be any kind of data file (text,
//      binary, executable, whatever).  Passwords can be up to 30 chars long.
//   3) Supply a destination file name.  This can be the same as the source
//      file name if you wish to overwrite the source file.
//   4) Optional: Set the temporary directory.  The default is the windows
//      system temporary directory.  If this is unacceptable, you may change
//      it to whatever you like.
//   5) Optional: Set KeyValidation to True or False.  If True, the password(s)
//      you enter will be encrypted and stored in the output file.  This way,
//      Emerald will be able to tell you if you enter an invalid password later
//      when trying to decrypt the file.  Setting it to False will prevent
//      Emerald from writing the password to the output file.  This provides
//      better security.
//   6) Optional: Set progress bar(s).  If you do, they will be updated
//      during encryption/decryption.  The "minor" progress bar is for
//      encryption with the current key.  The "major" progress bar is for
//      key-to-key encryption.  If you only want one, use the "major" one.
//      If your application will deal with very large files, set both.
//      If you don't set either, you will have no visual indication
//      of how the operation is going.
//   7) Call the Encrypt or Decrypt method.  Encryption/decryption will take place.
//
// NOTES:
//     You must know the password to decrypt a file.  There is no known way
//   to decrypt an Emerald (c) encrypted file without the original password.
//
// DISK SPACE:
//     It is important to note that you must have at least twice the space of
//   the source file free on your drive in order for the routine to function
//   correctly.  For example, if you wish to encrypt a 100 MB file, you must
//   have at least 200 MB free on your drive in order to do this.  This is
//   due to the multi-stage encryption, which requires two copies of the file to
//   exist simultaneously.  After encryption, any temp file created will be
//   deleted.  This should only be a problem on very small drives or when
//   encrypting very large files.
//
// REVISION HISTORY:
//   version 1.0 - implemented ASCII "shift by one" method (extremely simplistic)
//   version 2.0 - implemented substitution cypher method (fairly simplistic)
//   version 3.0 - improved user interface, same method as version 2
//   version 4.0 - key-encryption using bit rotation (average encryption)
//   version 5.0 - multiple key byte reversal encryption using bit rotation (fairly advanced)
//   version 6.0 - multiple key byte reversal encryption using key XOR (sufficiently advanced)
//   version 6.1 - Repaired a minor glitch in the byte reversal scheme
//   version 6.2 - added password validation
//   version 6.3 - made password validation optional, added footer for forwards compatibility
//   version 6.4 - fixed minor registry error
//
// November 30 1998
// Updated January 11 1999
// Steve Corbett
// steve@secretportal.com
// http://www.secretportal.com/Emerald/
//
////////////////////////////////////////////////////////////////////////////////

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, comctrls;

const

  // This is the version string that will be written into the file footer:
  CurrentEmeraldVersion = 'v6.40';

  // One of these flags will be written to the footer to signal whether
  // key validation was used or not:
  KeyValidationPositive = 'kvp';
  KeyValidationNegative = 'kvn';

  // This ident array is an arbitrary string of numbers that will be written
  // to the footer to identify the file as being an Emerald file.
  IdentArrayLength = 12;
  EmeraldIdentArray : Array[1..IdentArrayLength] of Byte = (42,11,99,52,23,16,12,8,90,11,42,42);

  // The size of the footer is computed as follows: five bytes for the version
  // string, three bytes for the key validation flag, and twelve bytes for the
  // ident string for a total of 5+3+12 = 20 bytes.
  EmeraldFooterSize = 20;

  // For keyfile generation, if you don't want to choose how many lines,
  // just pass in this constant and the Emerald object will choose for you.
  AutomaticSelection = -1;

  // Max key size.  Each individual key cannot exceed this length, in bytes:
  MaxKeySize = 99;

  // This is our file transfer buffer size.  Setting this to a lower
  // value uses less memory but takes a v-e-r-y long time on large
  // files.  Increasing the size of the buffer decreases operation time.
  TransferBufferSize = 32768; // 32K at a time sounds about right.

  // Some typical error messages:
  EmeraldErrMsg_KeyTooLong = 'A key that you entered exceeded the maximum key size (99).';
  EmeraldErrMsg_NoKey = 'You must specify a password in order to encrypt/decrypt.';
  EmeraldErrMsg_NoSource = 'You must specify a source file name.';
  EmeraldErrMsg_NoDest = 'You must specify a destination file name.';
  EmeraldErrMsg_BadSource = 'Unable to open input file!';
  EmeraldErrMsg_BadDest = 'Unable to create destination file!';
  EmeraldErrMsg_BadPassword = 'The password you entered does not match the one in the file.';
  EmeraldErrMsg_BadKeyFile = 'Unable to open the key file!';
  EmeraldErrMsg_CantCreateKeyFile = 'Unable to create the key file!';

type
  EmeraldError = class(Exception);

  TEmerald = class(TComponent)
  private
    // Store the source and destination file names:
    fInFileName, fOutFileName: String;

    // Temporary directory.  User-adjustable.
    fTemporaryDir: String;

    // Instead of typing in a password, input can come from a file:
    fKeyFileName: String;

    // Toggles whether the key (encrypted) is placed at the end of the file.
    // This allows Emerald to inform you when you have entered an invalid
    // password when decrypting a file, but it does somewhat compromise
    // security.  For maximum security, set this to FALSE.  For maximum
    // user-friendliness, set this to TRUE.
    fKeyValidation : Boolean;

    // The "raw" user key is stored here.
    fRawPassword: String;

    // The raw key is parsed (space delimited) into an array of separate strings:
    fPasswords: TStringList;

    // Data is transferred at a rate of up to 32 Kilobytes at a time:
    fTransferBuffer: Array[0..(TransferBufferSize-1)] of Byte;

    // Calling process can optionally set these if they want:
    fMinorProgressBar: TProgressBar;
    fMajorProgressBar: TProgressBar;

    // Called to parse the raw user key into separate strings:
    procedure ParseUserKey;

    // Getters/setters:
    procedure SetRawPassword(str: String);
    procedure SetKeyFile(str: String);

    // Called by both "Encrypt" and "Decrypt" to
    // perform common validity checks and to open both files:
    procedure CommonInitialization;

    // Support methods called by the main "Encrypt" and "Decrypt" methods:
    procedure EncryptFileToFile(SourceFileName, DestFileName, Key: String);
    procedure DecryptFileToFile(SourceFileName, DestFileName, Key: String);

    // Appends the encrypted raw password to the end of the specified file:
    procedure AppendEncryptedKey(filename: String);

    // Removes the encrypted raw password from the end of the specified file:
    procedure RemoveEncryptedKey(filename: String);

    // Appends the footer to the encrypted file:
    procedure AppendEmeraldFooter(filename: String);

    // File management methods:
    procedure RenameFile(oldName, newName: String);
    procedure CopyFile(oldLocation, newLocation: String);
    procedure DeleteFile(killfile: String);

  public
    // To encrypt, set properties and call this method:
    procedure Encrypt;

    // To decrypt, call this method.  Setting ValidatePassword to True will
    // raise an exception if the password fails validity check.  To decrypt
    // without checking the password, set ValidatePassword to False.
    procedure Decrypt(ValidatePassword: Boolean);

    // To determine if the file named in InFileName is actually an Emerald
    // encrypted file, call this method:
    function IsEmeraldEncrypted : Boolean;

    // To determine if key validation was used or not, call this method.
    // This method calls IsEmeraldEncrypted to ensure that the file
    // is an actual Emerald file.  File name is taken from InFileName.
    function KeyValidationWasUsed : Boolean;

    // To get the version of Emerald that was used to encrypt the file named
    // in InFileName, call this method.  This method will call the IsEmeraldEncrypted
    // method to ensure that the file is an actual Emerald file.  If no version
    // string is found or the file is not Emerald encrypted, an empty string
    // is returned.  Otherwise, the string is in the following
    // format: 'Emerald Encryption vX.XX' where x is the version number.
    function EmeraldVersionUsed : String;

    // Tries to match given password with encrypted file's password.
    // Returns TRUE if they match, FALSE if not.
    // File name is taken from InFileName property.
    function KeyIsValid: Boolean;

    // Called to generate a key file.  A key file is lots
    // of lines of random ASCII characters that can be used as a password
    // later for maximum security (more secure than typing in an English
    // word for a password).
    procedure GenerateKeyFile(filename: String; howmanyLines: Integer);

    // Standard stuff:
    constructor Create(owner: TComponent); override;
    destructor Destroy; override;

  published

    // "Published" properties are the ones that will show up in the Delphi
    // object inspector and can be set at design time.
    property InFileName: String read fInFileName write fInFileName;
    property OutFileName: String read fOutFileName write fOutFileName;
    property Password: String read fRawPassword write SetRawPassword;
    property MinorProgressBar: TProgressBar read fMinorProgressBar write fMinorProgressBar;
    property MajorProgressBar: TProgressBar read fMajorProgressBar write fMajorProgressBar;
    property TemporaryDir: String read fTemporaryDir write fTemporaryDir;
    property KeyFile: String read fKeyFileName write SetKeyFile;
    property KeyValidation: Boolean read fKeyValidation write fKeyValidation;
  end;

procedure Register;

implementation

procedure Register;
begin
  // This routine is called when the component is first installed.
  RegisterComponents('Emerald', [TEmerald]);
end;


constructor TEmerald.Create(owner: TComponent);
var
  tempstr: pChar;
begin
  // Call inherited:
  inherited Create(owner);

  // Default values:
  fMinorProgressBar := nil;
  fMajorProgressBar := nil;
  fInFileName := '';
  fOutFileName := '';
  fRawPassword := '';
  fKeyValidation := True;

  // Get system temporary directory:
  tempstr := StrAlloc(512);
  if GetTempPath(512,tempstr) = 0 then
    fTemporaryDir := 'C:\' // default
  else
    fTemporaryDir := StrPas(tempstr); // usually c:\windows\temp\
  StrDispose(tempstr);

  // Make sure it ends with a backslash:
  if fTemporaryDir[Length(fTemporaryDir)] <> '\' then
    fTemporaryDir := fTemporaryDir + '\';

  // Allocate memory:
  fPasswords := TStringList.Create;
end;

destructor TEmerald.Destroy;
begin
  // Release memory:
  fPasswords.Free;
  inherited;
end;

procedure TEmerald.ParseUserKey;
var
  TempStr: String;
  str: String;
  finished: Boolean;
begin
  // Clear out old list:
  fPasswords.Clear;

  // Remove leading and trailing spaces:
  fRawPassword := Trim(fRawPassword);

  // if there are no spaces in the password, take it as-is:
  if Pos(' ',fRawPassword)=0 then
  begin
    str := fRawPassword;
    if Length(str) > MaxKeySize then
      raise EmeraldError.Create(EmeraldErrMsg_KeyTooLong);
    fPasswords.Add(str);
  end

  // If there are spaces, separate each word out:
  else begin
    TempStr := fRawPassword;
    finished := False;
    repeat
      // Get the first word:
      str := Copy(TempStr,1,Pos(' ',TempStr)-1);

      // Now remove the first word from the raw password:
      TempStr := Copy(TempStr,Pos(' ',TempStr)+1,Length(TempStr));

      // Trim the remaining raw password (removes leading spaces).
      // This prevents a serious error from occurring.  I discovered
      // this by accident when I entered two words separated by two
      // spaces instead of just one.
      TempStr := Trim(TempStr);

      // Validity check:
      if Length(str)>MaxKeySize then
        raise EmeraldError.Create(EmeraldErrMsg_KeyTooLong);

      // Accept:
      fPasswords.Add(str);

      // Are there more spaces?
      if Pos(' ',TempStr)=0 then
      begin
        if Length(TempStr)>MaxKeySize then
          raise EmeraldError.Create(EmeraldErrMsg_KeyTooLong);

        fPasswords.Add(TempStr);

        Finished := True;
      end;

    until Finished;
  end;
end;

procedure TEmerald.SetRawPassword(str: String);
begin
  // Take the string and parse it into separate words.
  fRawPassword := str;
  ParseUserKey;
end;

procedure TEmerald.SetKeyFile(str: String);
var
  keyFile: Textfile;
  tempStr: String;
begin
  // accept:
  fKeyFileName := str;

  // Try to open:
  AssignFile(keyFile,fKeyFileName);
  try
    Reset(keyFile);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadKeyFile);
  end;

  // Clear out existing passwords:
  fRawPassword := '';
  fPasswords.Clear;

  // Read each line:
  repeat
    // read a line:
    ReadLn(keyFile,tempStr);

    // Truncate if necessary:
    if Length(tempStr) > MaxKeySize then
      tempStr := Copy(tempStr,1,MaxKeySize);

    // Accept:
    fRawPassword := fRawPassword + tempStr;
    fPasswords.Add(tempStr);

  until EOF(keyFile);

  // Close up:
  CloseFile(keyFile);
end;

procedure TEmerald.CommonInitialization;
begin
  // Validity check(s):
  if fPasswords.Count = 0 then
    raise EmeraldError.Create(EmeraldErrMsg_NoKey);
  if fInFileName = '' then
    raise EmeraldError.Create(EmeraldErrMsg_NoSource);
  if fOutFileName = '' then
    fOutFileName := fInFileName;
end;

procedure TEmerald.AppendEncryptedKey(filename: String);
var
  keyFile: File;
  eKey: TByteArray;
  ctr: Integer;
begin
  // Open specified file:
  AssignFile(keyFile,filename);
  try
    Reset(keyFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Move to end of file:
  Seek(keyFile,FileSize(keyFile));

  // Encrypt the raw user key:
  for ctr := 0 to Length(fRawPassword)-1 do
    eKey[ctr] := Byte(fRawPassword[ctr+1]) xor Byte(fRawPassword[Length(fRawPassword)-ctr]);

  // Write this out:
  BlockWrite(keyFile,eKey,Length(fRawPassword));

  // Close up:
  CloseFile(keyFile);
end;

procedure TEmerald.RemoveEncryptedKey(filename: String);
var
  keyFile: File;
  tempFile: File;
  NumRead: Integer;
  FilePosition,keyFileSize: LongInt;
  CutOff: LongInt;
  AmountToWrite: LongInt;
  KeyIsPresent: Boolean;
begin
  // Was key validation used on this file?
  KeyIsPresent := KeyValidationWasUsed;

  // Open file:
  AssignFile(keyFile,filename);
  try
    reset(keyFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Open temp file:
  AssignFile(tempFile,fTemporaryDir+'eTempEMR.tmp');
  try
    rewrite(tempFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadDest);
  end;

  // Copy contents of source file to temp file, leaving out the key and the footer:
  FilePosition := 0;
  keyFileSize := FileSize(KeyFile);
  if KeyIsPresent then
    CutOff := keyFileSize - Length(fRawPassword) - EmeraldFooterSize
  else
    CutOff := KeyFileSize - EmeraldFooterSize;
  repeat
    // Get some data:
    BlockRead(keyFile,fTransferBuffer,TransferBufferSize,NumRead);

    // Are we there yet?
    if (filePosition + NumRead) > CutOff then
      AmountToWrite := NumRead - (FilePosition + NumRead - CutOff)
    else
      AmountToWrite := NumRead;

    // Update:
    FilePosition := FilePosition + NumRead;

    // Write out:
    BlockWrite(tempFile,fTransferBuffer,AmountToWrite);

  until (NumRead = 0) or (FilePosition >= CutOff);

  // Close up:
  CloseFile(keyFile);
  CloseFile(tempFile);
  RenameFile(fTemporaryDir+'eTempEMR.tmp',filename);
end;

procedure TEmerald.AppendEmeraldFooter(filename: String);
var
  f: File;
  ctr: Integer;
  footer: Array[1..EmeraldFooterSize] of Byte;
begin
  // The first thing in the footer is the Emerald Version.
  for ctr := 1 to 5 do
    footer[ctr] := Byte(CurrentEmeraldVersion[ctr]);

  // Next is the key validation positive/negative string:
  for ctr := 1 to 3 do
    if KeyValidation then
      footer[ctr+5] := Byte(KeyValidationPositive[ctr])
    else
      footer[ctr+5] := Byte(KeyValidationNegative[ctr]);

  // Finally, the ident array:
  for ctr := 1 to IdentArrayLength do
    footer[ctr+8] := EmeraldIdentArray[ctr];

  // Open the file:
  AssignFile(f,filename);
  try
    Reset(f,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadDest);
  end;

  // Skip to end:
  Seek(f,FileSize(f));

  // Write footer:
  BlockWrite(f,footer,EmeraldFooterSize);

  // Close up:
  CloseFile(f);
end;

function TEmerald.KeyIsValid: Boolean;
var
  numread: Integer;
  testFile: File;
  fileKey: TByteArray;
  dKey: TByteArray;
  ctr: Integer;
  Match: Boolean;
begin
  // Test:
  if fPasswords.Count = 0 then
    raise EmeraldError.Create(EmeraldErrMsg_NoKey);
  if fInFileName = '' then
    raise EmeraldError.Create(EmeraldErrMsg_NoSource);

  // Open:
  AssignFile(testFile,fInFileName);
  try
    Reset(testFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // If the password is longer than the file, forget it:
  if FileSize(testFile) < Length(fRawPassword) then
  begin
    KeyIsValid := False;
    CloseFile(testFile);
    Exit;
  end;

  // Seek to the start of the key:
  Seek(testFile,FileSize(testFile)-EmeraldFooterSize-Length(fRawPassword));

  // Read it in:
  BlockRead(testFile,fileKey,Length(fRawPassword),NumRead);

  // Check:
  if NumRead <> Length(fRawPassword) then
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);

  // Close off:
  CloseFile(testFile);

  // Decrypt it:
  for ctr := 0 to Length(fRawPassword)-1 do
    dKey[ctr] := fileKey[ctr] xor Byte(fRawPassword[Length(fRawPassword)-ctr]);

  // Do they match?
  Match := True;
  for ctr := 0 to Length(fRawPassword)-1 do
    if Byte(fRawPassword[ctr+1]) <> dKey[ctr] then
      Match := False;

  // Report results
  KeyIsValid := Match;
end;

procedure TEmerald.EncryptFileToFile(SourceFileName, DestFileName, Key: String);
var
  SourceFile, DestFile: File;
  NumRead: Integer;
  KeyLength: Integer;
  KeyCtr: Integer;
  TempHoldSize,ctr1,ctr2: Integer;
  CurrentByte: Byte;
  TempHold: Array[0..MaxKeySize] of Byte;
  fBufferPosition: Integer;
  fCurrentFilePosition,fTotalFileSize: LongInt;
begin
  // Open source file:
  AssignFile(SourceFile,SourceFileName);
  try
    Reset(SourceFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Get length of source file:
  fTotalFileSize := FileSize(SourceFile);
  fCurrentFilePosition := 0;

  // Create destination file:
  AssignFile(DestFile,fTemporaryDir+DestFileName);
  try
    Rewrite(DestFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadDest);
  end;

  // Set up minor progress bar if assigned:
  if fMinorProgressBar <> nil then
  begin
    fMinorProgressBar.Max := 100;
    fMinorProgressBar.Min := 0;
    fMinorProgressBar.Position := 0;
  end;

  // Go:
  repeat
    // Make a note of the length of this password:
    KeyLength := Length(Key);

    // Get some data:
    BlockRead(SourceFile,fTransferBuffer,TransferBufferSize,NumRead);

    // Go through the buffer:
    fBufferPosition := 0;
    repeat

      // Get KEYLENGTH or less bytes of data out of the transfer buffer:
      TempHoldSize := 0;
      if NumRead > 0 then repeat
        TempHold[TempHoldSize] := fTransferBuffer[fBufferPosition+TempHoldSize];
        TempHoldSize := TempHoldSize+1;
      until (TempHoldSize >= KeyLength) or (fBufferPosition+TempHoldSize >= NumRead);

      // Loop to encrypt each Byte value:
      KeyCtr := 1;
      for ctr2 := 0 to TempHoldSize-1 do
      begin
        // XOR this byte with a byte from the key:
        Currentbyte := TempHold[ctr2] xor Byte(Key[KeyCtr]);

        // Put it back into the holding buffer, encrypted.
        TempHold[ctr2] := CurrentByte;

        // Keep looping through the bytes in the key:
        Inc(KeyCtr);
      end;

      // Loop to reverse the hold buffer in memory (neat o mosquito).
      // This is the main feature that sets Emerald apart from other encryption
      // methods.  Not only is each byte in the input file encrypted, but
      // the order of the source bytes is also messed up in a controlled way.
      ctr1 := 0;
      ctr2 := TempHoldSize-1;
      while Ctr2 > ctr1 do
      begin
        // Swap.
        CurrentByte := TempHold[ctr1];
        TempHold[ctr1] := TempHold[ctr2];
        TempHold[ctr2] := CurrentByte;
        Inc(Ctr1);
        Dec(Ctr2);
      end;

      // Put this back into the transfer buffer:
      for ctr1 := fBufferPosition to fBufferPosition + TempHoldSize-1 do
        fTransferBuffer[ctr1] := TempHold[ctr1-fBufferPosition];

      // Adjust:
      fBufferPosition := fBufferPosition + TempHoldSize;

    until fBufferPosition >= NumRead;

    // Write this out:
    BlockWrite(DestFile,fTransferBuffer,NumRead);

    // Update current position:
    fCurrentFilePosition := fCurrentFilePosition + NumRead;

    // Update minor progress bar if assigned:
    if fMinorProgressBar <> nil then
      fMinorProgressBar.Position := Round((fCurrentFilePosition / fTotalFileSize) * 100);

    // Allow application to remain responsive.
    // If you remove this line, you get "Microsoft" style behaviour, where the
    // application completely stops responding to input while it is working.
    // To see why I call this "Microsoft" behaviour, take a look at MS ACCESS.
    // NOTE: this does slow down the process a bit.
    Application.ProcessMessages;

  until (NumRead = 0) or EOF(SourceFile);

  // Close up:
  CloseFile(SourceFile);
  CloseFile(DestFile);
end;

procedure TEmerald.DecryptFileToFile(SourceFileName, DestFileName, Key: String);
var
  SourceFile, DestFile: File;
  NumRead: Integer;
  KeyLength: Integer;
  TempHoldSize,ctr1,ctr2: Integer;
  KeyCtr: Integer;
  CurrentByte: Byte;
  fBufferPosition: Integer;
  TempHold: Array[0..MaxKeySize] of Byte;
  fCurrentFilePosition,fTotalFileSize: LongInt;
begin
  // Open source file:
  AssignFile(SourceFile,SourceFileName);
  try
    Reset(SourceFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Get length of source file:
  fTotalFileSize := FileSize(SourceFile);
  fCurrentFilePosition := 0;

  // Create destination file:
  AssignFile(DestFile,fTemporaryDir+DestFileName);
  try
    Rewrite(DestFile,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadDest);
  end;

  // Set up minor progress bar if assigned:
  if fMinorProgressBar <> nil then
  begin
    fMinorProgressBar.Max := 100;
    fMinorProgressBar.Min := 0;
    fMinorProgressBar.Position := 0;
  end;

  // Go:
  repeat
    // Make a note of the length of this password:
    KeyLength := Length(Key);

    // Get some data:
    BlockRead(SourceFile,fTransferBuffer,TransferBufferSize,NumRead);

    // Go through the buffer:
    fBufferPosition := 0;
    repeat

      // Get KEYLENGTH or less bytes of data out of the transfer buffer:
      TempHoldSize := 0;
      if NumRead > 0 then repeat
        TempHold[TempHoldSize] := fTransferBuffer[fBufferPosition+TempHoldSize];
        TempHoldSize := TempHoldSize+1;
      until (TempHoldSize >= KeyLength) or (fBufferPosition+TempHoldSize >= NumRead);

      // Restore byte order before decrypting:
      ctr1 := 0;
      ctr2 := TempHoldSize-1;
      while Ctr2 > ctr1 do
      begin
        // Swap.
        CurrentByte := TempHold[ctr1];
        TempHold[ctr1] := TempHold[ctr2];
        TempHold[ctr2] := CurrentByte;
        Inc(Ctr1);
        Dec(Ctr2);
      end;

      // Loop to decrypt each Byte value:
      KeyCtr := 1;
      for ctr2 := 0 to TempHoldSize-1 do
      begin
        // XOR this byte with a byte from the key:
        Currentbyte := TempHold[ctr2] xor Byte(Key[KeyCtr]);

        // Put it back into the holding buffer, encrypted.
        TempHold[ctr2] := CurrentByte;

        // Keep looping through the bytes in the key:
        Inc(KeyCtr);
      end;

      // Put this back into the transfer buffer:
      for ctr1 := fBufferPosition to fBufferPosition + TempHoldSize-1 do
        fTransferBuffer[ctr1] := TempHold[ctr1-fBufferPosition];

      // Adjust:
      fBufferPosition := fBufferPosition + TempHoldSize;

    until fBufferPosition >= NumRead;

    // Write this out:
    BlockWrite(DestFile,fTransferBuffer,NumRead);

    // Update current position:
    fCurrentFilePosition := fCurrentFilePosition + NumRead;

    // Update minor progress bar if assigned:
    if fMinorProgressBar <> nil then
      fMinorProgressBar.Position := Round((fCurrentFilePosition / fTotalFileSize) * 100);

    // Allow application to remain responsive.
    // If you remove this line, you get "Microsoft" style behaviour, where the
    // application completely stops responding to input while it is working.
    // To see why I call this "Microsoft" behaviour, take a look at MS ACCESS.
    // NOTE: this does slow down the process a bit.
    Application.ProcessMessages;

  until (NumRead = 0) or EOF(SourceFile);

  // Close up:
  CloseFile(SourceFile);
  CloseFile(DestFile);
end;

procedure TEmerald.RenameFile(oldName, newName: String);
var
  f: File;
begin
  // First, we simply try to rename the file.  If a file with the new name
  // already exists, this will fail.
  try
    AssignFile(f,oldName);
    Rename(f,newName);
  except
    // If it failed, it could be because a file with this new name already
    // exists, so let's try deleting it.
    DeleteFile(newName);
    try
      // Try again.
      Rename(f,newName);
    except
      // If that still didn't work, abort.
      // This could be because the new file name points to an invalid path
      // or to a drive that is read only (cd drive, network drive).
      // There is nothing we can do at this point.
      raise EmeraldError.Create(EmeraldErrMsg_BadDest);
    end;
  end;
end;

procedure TEmerald.GenerateKeyFile(filename: String; howmanyLines: Integer);
const
  ValidCharacters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz!@#$%^&*()_-[]{};<>?/\|';
var
  keyFile: TextFile;
  ctr: Integer;
  TempStr: String;
  NumberOfLines: Integer;
  LineCtr: Integer;
begin
  Randomize;

  // Open:
  AssignFile(keyFile,filename);
  try
    rewrite(keyfile);
  except
    raise EmeraldError.Create(EmeraldErrMsg_CantCreateKeyFile);
  end;

  // Did they tell us how many lines to generate?
  if howManyLines <> AutomaticSelection then
    NumberOfLines := HowManyLines

  // Otherwise, let's just arbitrarily choose:
  else
    NumberOfLines := 20; // lots

  // Make this many lines:
  LineCtr := 1;
  repeat
    // Create a line of gibberish:
    TempStr := '';
    for ctr := 1 to (5+Random(MaxKeySize-5)) do // at least five characters
      TempStr := TempStr + ValidCharacters[Random(Length(ValidCharacters)-1)+1];

    // Write out this line:
    WriteLn(keyFile,TempStr);

    Inc(LineCtr);

  until LineCtr > NumberOfLines;

  // Close:
  CloseFile(keyFile);
end;

procedure TEmerald.CopyFile(oldLocation,newLocation: String);
var
  inF,outF: File;
  NumRead: Integer;
begin
  // Open source file:
  AssignFile(inF,oldLocation);
  try
    Reset(inF,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Open dest file:
  AssignFile(outF,newLocation);
  Rewrite(outF,1);

  // Copy contents:
  repeat
    BlockRead(inF,fTransferBuffer,TransferBufferSize,NumRead);
    BlockWrite(outF,fTransferBuffer,NumRead);
  until (EOF(inF)) or (NumRead = 0);

  // Close:
  CloseFile(inF);
  CloseFile(outF);
end;

procedure TEmerald.DeleteFile(killfile: String);
var
  f: File;
begin
  // Delete specified file.
  AssignFile(f,killfile);
  Erase(f);
end;

procedure TEmerald.Encrypt;
var
  CurrentKey: Integer;
  SourceName,DestName: String;
  TrashFile: String;
begin
  // Get the mundane stuff out of the way:
  CommonInitialization;
  SourceName := fInFileName;
  DestName := 'EETemp0.emr';
  TrashFile := '';

  // Start with the first key:
  CurrentKey := 0;

  // Set up the major progress bar if there is one:
  if fMajorProgressBar <> nil then
  begin
    fMajorProgressBar.Max := fPasswords.Count-1;
    fMajorProgressBar.Min := 0;
    fMajorProgressBar.Position := 0;
  end;

  // Encrypt the file once for every space delimited key:
  repeat
    // Make one pass:
    EncryptFileToFile(SourceName,DestName,fPasswords[CurrentKey]);

    // Next key:
    Inc(CurrentKey);

    // Adjust file names:
    TrashFile := SourceName;
    SourceName := fTemporaryDir+DestName;
    DestName := 'EETemp'+IntToStr(CurrentKey)+'.emr';

    // Delete temp files as necessary:
    if TrashFile <> fInFilename then
      DeleteFile(TrashFile);

    // Update progress bar if necessary:
    if fMajorProgressBar <> nil then
      fMajorProgressBar.Position := fMajorProgressBar.Position + 1;

  until CurrentKey >= fPasswords.Count;

  // Rename the encrypted file to the destination name:
  RenameFile(SourceName,fOutFileName);

  // Append the encrypted key to the destination file:
  if KeyValidation then
    AppendEncryptedKey(fOutFileName);

  // Write out the emerald footer:
  AppendEmeraldFooter(fOutFileName);
end;

procedure TEmerald.Decrypt(ValidatePassword: Boolean);
var
  CurrentKey: Integer;
  SourceName,DestName: String;
  TrashFile: String;
begin
  // Get the mundane stuff out of the way:
  CommonInitialization;
  SourceName := fInFileName;
  DestName := 'EDTemp.emr';
  TrashFile := '';

  // Check password if instructed to do so:
  if ValidatePassword then
    if not KeyIsValid then
      raise EmeraldError.Create(EmeraldErrMsg_BadPassword);

  // Copy the file to our temporary directory so that we can
  // play with it:
  CopyFile(SourceName,DestName);
  SourceName := DestName;
  DestName := 'EDTemp0.emr';

  // Remove the encrypted key from this file:
  RemoveEncryptedKey(SourceName);

  // Start with the last key:
  CurrentKey := fPasswords.Count-1;

  // Set up the major progress bar if there is one:
  if fMajorProgressBar <> nil then
  begin
    fMajorProgressBar.Max := fPasswords.Count-1;
    fMajorProgressBar.Min := 0;
    fMajorProgressBar.Position := 0;
  end;

  repeat
    // Make one pass:
    DecryptFileToFile(SourceName,DestName,fPasswords[CurrentKey]);

    // Next key:
    Dec(CurrentKey);

    // Adjust file names:
    TrashFile := SourceName;
    SourceName := fTemporaryDir+DestName;
    DestName := 'EDTemp'+IntToStr(fPasswords.Count-1-CurrentKey)+'.emr';

    // Delete temporary files as necessary:
    if TrashFile <> fInFileName then
      DeleteFile(TrashFile);

    // Update progress bar if necessary:
    if fMajorProgressBar <> nil then
      fMajorProgressBar.Position := fMajorProgressBar.Position + 1;

  until CurrentKey < 0;

  // Rename the decrypted file to the destination name:
  RenameFile(SourceName,fOutFileName);
end;

function TEmerald.IsEmeraldEncrypted : Boolean;
var
  f: File;
  ctr: Integer;
  TestStr: String;
  buffer: Array[1..EmeraldFooterSize] of byte;
begin
  // Open the file:
  AssignFile(f,fInFileName);
  try
    Reset(f,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // If the file is smaller than the footer, we have a problem:
  if FileSize(f) < EmeraldFooterSize then
  begin
    CloseFile(f);
    IsEmeraldEncrypted := False;
    Exit;
  end;

  // Skip to the footer:
  Seek(f,FileSize(f)-EmeraldFooterSize);

  // Read in the footer:
  BlockRead(f,buffer,EmeraldFooterSize);

  // Close the file:
  CloseFile(f);

  // Skip the version string, since it could be anything.
  // Move to the key validation flag.  This must be either
  // KeyValidationPositive or KeyValidationNegative.
  TestStr := '';
  for ctr := 6 to 8 do
    TestStr := TestStr + char(buffer[ctr]);
  if (TestStr <> KeyValidationPositive) and (TestStr <> KeyValidationNegative) then
  begin
    IsEmeraldEncrypted := False;
    Exit;
  end;

  // Now check the ident array.  This must match our ident array for the file
  // to pass inspection.
  for ctr := 9 to EmeraldFooterSize do
    if buffer[ctr] <> EmeraldIdentArray[ctr-8] then
    begin
      IsEmeraldEncrypted := False;
      Exit;
    end;

  // If it passed all of the above tests, it must be good:
  IsEmeraldEncrypted := True;
end;

function TEmerald.KeyValidationWasUsed : Boolean;
var
  f: File;
  ctr: Integer;
  buffer: Array[1..3] of Byte;
  TestStr: String;
begin
  // First make sure the file is an actual Emerald file:
  if not IsEmeraldEncrypted then
  begin
    KeyValidationWasUsed := False;
    Exit;
  end;

  // Open the file:
  AssignFile(f,fInFileName);
  try
    Reset(f,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Skip to the key validation flag:
  Seek(f,FileSize(f) - IdentArrayLength - 3);

  // Read in the flag:
  BlockRead(f,buffer,3);

  // Close the file:
  CloseFile(f);

  // Move from byte array into string:
  TestStr := '';
  for ctr := 1 to 3 do
    TestStr := TestStr + char(buffer[ctr]);

  // Positive or negative?
  // We know it is one or the other because it couldn't have
  // passed the IsEmeraldEncrypted test otherwise.
  if TestStr = KeyValidationPositive then
    KeyValidationWasUsed := True
  else
    KeyValidationWasUsed := False;
end;

function TEmerald.EmeraldVersionUsed : String;
var
  f: File;
  ctr: Integer;
  buffer: Array[1..5] of byte;
  VersionStr: String;
begin
  // First make sure the file is an actual Emerald file:
  if not IsEmeraldEncrypted then
  begin
    EmeraldVersionUsed := '';
    Exit;
  end;

  // Open the file:
  AssignFile(f,fInFileName);
  try
    Reset(f,1);
  except
    raise EmeraldError.Create(EmeraldErrMsg_BadSource);
  end;

  // Skip to the version string:
  Seek(f,FileSize(f) - EmeraldFooterSize);

  // Read in the version string:
  BlockRead(f,buffer,5);

  // Transfer from byte array to string:
  VersionStr := '';
  for ctr := 1 to 5 do
    VersionStr := VersionStr + char(buffer[ctr]);

  // Close the file:
  CloseFile(f);

  // Return whatever we got out of that.  We can't do any validity checks
  // at this point because the version string can be anything.
  EmeraldVersionUsed := VersionStr;
end;


end.
