unit PersistentForm;

////////////////////////////////////////////////////////////////////////////////
//
// PersistentForm
//
// This is a quick and dirty class which extends the Delphi TForm object to
// allow a form to save its position to the registry at program close and load
// it when the program starts.  This allows the user to place the form wherever
// they want it on the screen and then have it come up that way again the next
// time they run the application.
//
// To keep things simple, I am only saving/loading the form's left and top
// properties, although this could be altered to save width and height as well.
//
////////////////////////////////////////////////////////////////////////////////

interface

uses
  Classes, Forms, Registry, Windows;

const
  cregDefaultKeyFormsRoot = '\Software\Secret Portal\Emerald' ;

type
  TFormPersistent = class(TForm)
  private
    Reg: TRegistry;
  public
    FormPersistenceKey : string ;
    SaveOnDestroy : boolean ;
    procedure Load ;
    procedure Save ;
    constructor Create (AOwner : TComponent) ; override ;
    destructor Destroy ; override ;
  end ;

implementation

uses
  SysUtils;


constructor TFormPersistent.Create (AOwner : TComponent) ;
begin
  //Initialize.
  Reg := TRegistry.Create;
  Reg.RootKey := HKEY_LOCAL_MACHINE;
  FormPersistenceKey := cregDefaultKeyFormsRoot ;
  SaveOnDestroy := True ;

  inherited Create (AOwner) ;
end ;


destructor TFormPersistent.Destroy ;
begin
  if SaveOnDestroy then
    Save ;

  Reg.Free;
  inherited Destroy ;
end ;


procedure TFormPersistent.Save ;
begin
  Reg.RootKey := HKEY_LOCAL_MACHINE;
  Reg.OpenKey(FormPersistenceKey,True);
  Reg.WriteInteger('Left',Left);
  Reg.WriteInteger('Top',Top);
  Reg.CloseKey;
end;


procedure TFormPersistent.Load ;
var
  testLeft, testTop: integer ;
begin
  // Default values:
  Left := (Screen.Width div 2) - (Width div 2);
  Top := (Screen.Height div 2) - (Height div 2);

  Reg.RootKey := HKEY_CURRENT_USER;
  if not Reg.OpenKey(FormPersistenceKey,False) then
  begin
    Save;
    Exit;
  end;

  try
    testLeft := Reg.ReadInteger('Left');
    testTop := Reg.ReadInteger('Top');

    //Enforce screen boundaries.
    if testLeft < 0 then
      testLeft := 0 ;
    if (testLeft + Width) > Screen.Width then
      testLeft := Screen.Width - Width ;
    if testTop < 0 then
      testTop := 0 ;
    if (testTop + Height) > Screen.Height then
      testTop := Screen.Height - Height;

    //Reposition and Resize.
    Left := testLeft ;
    Top := testTop ;
  except
    Save;
    Exit;
  end ;

  Reg.CloseKey;
end;


end.
